'''LSST-specific Github OAuthenticator class, delegating its LSST-specific
authentication to its auth_mgr.
'''
import json
import oauthenticator
from oauthenticator.common import next_page_from_links
from tornado import gen
from tornado.httpclient import HTTPRequest, AsyncHTTPClient, HTTPError
from .lsstauth import LSSTAuthenticator
from ..lsstmgr import check_membership
from ..utils import make_logger


def github_api_headers(access_token):
    '''Generate API headers for communicating with GitHub.
    '''
    return {"Accept": "application/json",
            "User-Agent": "JupyterHub",
            "Authorization": "token {}".format(access_token)
            }


class LSSTGitHubOAuthenticator(LSSTAuthenticator,
                               oauthenticator.GitHubOAuthenticator):
    login_handler = oauthenticator.OAuthLoginHandler
    scope = ['read:org', 'user:email', 'public_repo']

    def __init__(self, *args, **kwargs):
        self.log = make_logger()
        self.log.debug("Creating LSSTGitHubOAuthenticator.")
        super().__init__(*args, **kwargs)

    @gen.coroutine
    def authenticate(self, handler, data=None):
        '''Authenticate and store data on auth_mgr.
        '''
        self.log.info("Authenticating user against GitHub.")
        cfg = self.lsst_mgr.config
        # This must be set for the superclass authentication
        self.github_organization_whitelist = cfg.github_allowlist
        userdict = yield super().authenticate(handler, data)
        if userdict is None:
            return None
        ast = userdict["auth_state"]
        token = ast["access_token"]
        self.log.debug("Setting authenticator groups from token.")
        groupmap = yield self._get_github_user_organizations(token)
        groups = list(groupmap.keys())
        ok = check_membership(groups, cfg.allowed_groups, cfg.forbidden_groups,
                              log=self.log)
        if not ok:
            self.log.warning("Group membership check failed.")
            return None
        _ = yield self._set_github_user_email(ast, token)
        ast['group_map'] = groupmap
        ast['uid'] = ast['github_user']['id']
        return userdict

    @gen.coroutine
    def _get_github_user_organizations(self, access_token):
        # Requires 'read:org' token scope.
        http_client = AsyncHTTPClient()
        headers = github_api_headers(access_token)
        gh_api = self.lsst_mgr.config.github_api
        next_page = "https://{}/user/orgs".format(gh_api)
        orgmap = {}
        while next_page:
            req = HTTPRequest(next_page, method="GET", headers=headers)
            try:
                resp = yield http_client.fetch(req)
            except HTTPError as exc:
                self.log.error("{} -> {}".format(next_page, exc))
                return None
            resp_json = json.loads(resp.body.decode('utf8', 'replace'))
            next_page = next_page_from_links(resp)
            for entry in resp_json:
                group = entry["login"]
                orgmap[group] = entry["id"]
        return orgmap

    @gen.coroutine
    def _set_github_user_email(self, auth_state, access_token):
        gh_user = auth_state["github_user"]
        gh_email = gh_user.get("email")
        if gh_email:
            # Nothing further to do.
            return
        if not gh_email:
            gh_email = yield self._get_github_user_email(access_token)
        if not gh_email:
            # Oh well.
            return
        gh_user["email"] = gh_email
        # And now it will be stored with auth_state

    @gen.coroutine
    def _get_github_user_email(self, access_token):
        # Determine even private email, if the token has 'user:email'
        #  scope
        self.log.debug("Attempting alternate means of determining user email.")
        http_client = AsyncHTTPClient()
        headers = github_api_headers(access_token)
        gh_api = self.lsst_mgr.config.github_api
        next_page = "https://{}/user/emails".format(gh_api)
        while next_page:
            req = HTTPRequest(next_page, method="GET", headers=headers)
            resp = yield http_client.fetch(req)
            resp_json = json.loads(resp.body.decode('utf8', 'replace'))
            next_page = next_page_from_links(resp)
            for entry in resp_json:
                if "email" in entry:
                    if "primary" in entry and entry["primary"]:
                        self.log.debug(
                            "Found email: {}".format(entry['email']))
                        return entry["email"]
        self.log.debug("Could not determine user email.")
        return None



'''Class to provide namespace manipulation.
'''

import time
from kubernetes.client.rest import ApiException
from kubernetes import client
from .. import LoggableChild


class LSSTNamespaceManager(LoggableChild):
    '''Class to provide namespace manipulation.
    '''
    parent = None
    log = None
    namespace = None
    service_account = None  # Account for pod to run as

    def set_namespace(self, namespace):
        self.namespace = namespace

    def ensure_namespace(self):
        '''Here we make sure that the namespace exists, creating it if
        it does not.  That requires a ClusterRole that can list and create
        namespaces.

        If we create the namespace, we also create (if needed) a ServiceAccount
        within it to allow the user pod to spawn dask and workflow pods.

        '''
        namespace = self.namespace
        if not namespace or namespace == "default":
            raise ValueError("Will not use default namespace!")
        api = self.parent.api
        cfg = self.parent.config
        ns = client.V1Namespace(
            metadata=client.V1ObjectMeta(
                name=namespace,
                labels={'argocd.argoproj.io/instance': 'nublado-users'},
                annotations={
                    'argocd.argoproj.io/compare-options': 'IgnoreExtraneous',
                    'argocd.argoproj.io/sync-options': 'Prune=false',
                }))
        try:
            self.log.info("Attempting to create namespace '%s'" % namespace)
            api.create_namespace(ns)
        except ApiException as e:
            if e.status != 409:
                estr = "Create namespace '%s' failed: %s" % (ns, str(e))
                self.log.exception(estr)
                raise
            else:
                self.log.info("Namespace '%s' already exists." % namespace)
        # Wait for the namespace to actually appear before creating objects
        #  in it.
        self.wait_for_namespace()
        if cfg.allow_dask_spawn:
            self.log.debug("Ensuring namespaced service account.")
            self.ensure_namespaced_service_account()
        if self.parent.spawner.enable_namespace_quotas:
            # By the time we need this, quota will have been set, because
            #  we needed it for options form generation.
            self.log.debug("Determining resource quota.")
            qm = self.parent.quota_mgr
            qm.ensure_namespaced_resource_quota(qm.quota)
        self.log.debug("Namespace resources ensured.")

    def define_namespaced_account_objects(self):
        namespace = self.namespace
        username = self.parent.user.escaped_name
        account = "{}-svcacct".format(username)
        self.service_account = account
        md = client.V1ObjectMeta(
            name=account,
            labels={'argocd.argoproj.io/instance': 'nublado-users'},
            annotations={
                'argocd.argoproj.io/compare-options': 'IgnoreExtraneous',
                'argocd.argoproj.io/sync-options': 'Prune=false',
            }
        )
        svcacct = client.V1ServiceAccount(metadata=md)
        # These rules let us manipulate Dask pods, Argo Workflows, and
        #  Multus CNI interfaces
        rules = [
            client.V1PolicyRule(
                api_groups=["argoproj.io"],
                resources=["workflows", "workflows/finalizers"],
                verbs=["get", "list", "watch", "update", "patch", "create",
                       "delete"]
            ),
            client.V1PolicyRule(
                api_groups=["argoproj.io"],
                resources=["workflowtemplates",
                           "workflowtemplates/finalizers"],
                verbs=["get", "list", "watch"],
            ),
            client.V1PolicyRule(
                api_groups=[""],
                resources=["secrets"],
                verbs=["get"]
            ),
            client.V1PolicyRule(
                api_groups=[""],
                resources=["pods", "pods/exec", "services", "configmaps"],
                verbs=["get", "list", "watch", "create", "delete",
                       "update", "patch"]
            ),
            client.V1PolicyRule(
                api_groups=[""],
                resources=["pods/log", "serviceaccounts"],
                verbs=["get", "list", "watch"]
            ),
        ]
        role = client.V1Role(
            rules=rules,
            metadata=md)
        rolebinding = client.V1RoleBinding(
            metadata=md,
            role_ref=client.V1RoleRef(api_group="rbac.authorization.k8s.io",
                                      kind="Role",
                                      name=account),
            subjects=[client.V1Subject(
                kind="ServiceAccount",
                name=account,
                namespace=namespace)]
        )
        return svcacct, role, rolebinding

    def ensure_namespaced_service_account(self):
        # Create a service account with role and rolebinding to allow it
        #  to manipulate resources in the namespace.
        self.log.info("Ensuring namespaced service account.")
        namespace = self.namespace
        api = self.parent.api
        rbac_api = self.parent.rbac_api
        svcacct, role, rolebinding = self.define_namespaced_account_objects()
        account = self.service_account
        try:
            self.log.info("Attempting to create service account.")
            api.create_namespaced_service_account(
                namespace=namespace,
                body=svcacct)
        except ApiException as e:
            if e.status != 409:
                self.log.exception("Create service account '%s' " % account +
                                   "in namespace '%s' " % namespace +
                                   "failed: %s" % str(e))
                raise
            else:
                self.log.info("Service account '%s' " % account +
                              "in namespace '%s' already exists." % namespace)
        try:
            self.log.info("Attempting to create role in namespace.")
            rbac_api.create_namespaced_role(
                namespace,
                role)
        except ApiException as e:
            if e.status != 409:
                self.log.exception("Create role '%s' " % account +
                                   "in namespace '%s' " % namespace +
                                   "failed: %s" % str(e))
                raise
            else:
                self.log.info("Role '%s' " % account +
                              "already exists in namespace '%s'." % namespace)
        try:
            self.log.info("Attempting to create rolebinding in namespace.")
            rbac_api.create_namespaced_role_binding(
                namespace,
                rolebinding)
        except ApiException as e:
            if e.status != 409:
                self.log.exception("Create rolebinding '%s'" % account +
                                   "in namespace '%s' " % namespace +
                                   "failed: %s", str(e))
                raise
            else:
                self.log.info("Rolebinding '%s' " % account +
                              "already exists in '%s'." % namespace)

    def wait_for_namespace(self, timeout=30):
        '''Wait for namespace to be created.'''
        namespace = self.namespace
        for dl in range(timeout):
            self.log.debug("Checking for namespace " +
                           "{} [{}/{}]".format(namespace, dl, timeout))
            nl = self.parent.api.list_namespace(timeout_seconds=1)
            for ns in nl.items:
                nsname = ns.metadata.name
                if nsname == namespace:
                    self.log.debug("Namespace {} found.".format(namespace))
                    return
                self.log.debug(
                    "Namespace {} not present yet.".format(namespace))
            time.sleep(1)
        raise RuntimeError(
            "Namespace '{}' was not created in {} seconds!".format(namespace,
                                                                   timeout))

    def maybe_delete_namespace(self):
        '''Here we try to delete the namespace.  If it has no non-dask
        running pods, and it's not the default namespace, we can delete it."

        This requires a cluster role that can delete namespaces.'''
        self.log.debug("Attempting to delete namespace.")
        namespace = self.namespace
        if not namespace or namespace == "default":
            raise RuntimeError("Cannot delete default namespace!")
            return
        podlist = self.parent.api.list_namespaced_pod(namespace)
        clear_to_delete = True
        if podlist and podlist.items:
            clear_to_delete = self._check_pods(podlist.items)
        if not clear_to_delete:
            self.log.info("Not deleting namespace '%s'" % namespace)
            return False
        self.log.info("Deleting namespace '%s'" % namespace)
        self.parent.api.delete_namespace(namespace)
        return True

    def _check_pods(self, items):
        namespace = self.namespace
        for i in items:
            if i and i.status:
                phase = i.status.phase
                if (phase == "Running" or phase == "Unknown"
                        or phase == "Pending"):
                    pname = i.metadata.name
                    if pname.startswith("dask-"):
                        self.log.debug(
                            ("Abandoned dask pod '{}' can be " +
                             "reaped.").format(pname))
                        # We can murder abandoned dask pods
                        continue
                    self.log.warning(("Pod in state '{}'; cannot delete " +
                                      "namespace '{}'.").format(phase,
                                                                namespace))
                    return False
        # FIXME check on workflows as well.
        return True

    def destroy_namespaced_resource_quota(self):
        '''Remove quotas from namespace.
        You don't usually have to call this, since it will get
        cleaned up as part of namespace deletion.
        '''
        namespace = self.get_user_namespace()
        api = self.parent.api
        qname = "quota-" + namespace
        dopts = client.V1DeleteOptions()
        self.log.info("Deleting resourcequota '%s'" % qname)
        api.delete_namespaced_resource_quota(qname, namespace, dopts)

    def delete_namespaced_service_account_objects(self):
        '''Remove service accounts, roles, and rolebindings from namespace.
        You don't usually have to call this, since they will get
         cleaned up as part of namespace deletion.
        '''
        namespace = self.namespace
        account = self.service_account
        if not account:
            self.log.info("Service account not defined.")
            return
        dopts = client.V1DeleteOptions()
        self.log.info("Deleting service accounts/role/rolebinding " +
                      "for %s" % namespace)
        self.parent.rbac_api.delete_namespaced_role_binding(
            account,
            namespace,
            dopts)
        self.parent.rbac_api.delete_namespaced_role(
            account,
            namespace,
            dopts)
        self.parent.api.delete_namespaced_service_account(
            account,
            namespace,
            dopts)

    def dump(self):
        '''Return dict for pretty-printing.
        '''
        nd = {"namespace": self.namespace,
              "service_account": self.service_account,
              "parent": str(self.parent)}
        return nd



'''This is a JupyterHub KubeSpawner, extended with the ability to manipulate
namespaces, and with an lsst_mgr attribute.
'''
import json
from .multispawner import MultiNamespacedKubeSpawner
from kubespawner.objects import make_pod
from tornado import gen
from traitlets import Bool
from ..utils import make_logger, sanitize_dict


class LSSTSpawner(MultiNamespacedKubeSpawner):
    '''This, plus the LSST Manager class structure, implements the
    LSST-specific parts of our spawning requirements.
    '''
    lsst_mgr = None
    wf_api = None
    delete_grace_period = 5
    # In our LSST setup, there is a "provisionator" user, uid/gid 769,
    #  that is who we should start as.
    uid = 769
    gid = 769
    # The fields need to be defined; we don't use them.
    fs_gid = None
    supplemental_gids = []
    extra_labels = {}
    extra_annotations = []
    image_pull_secrets = None
    privileged = False
    working_dir = None
    lifecycle_hooks = {}  # This one will be useful someday.
    init_containers = []
    lab_service_account = None
    extra_container_config = None
    extra_pod_config = None
    extra_containers = []

    delete_namespace_on_stop = Bool(
        True,
        config=True,
        help='''
        If True, the entire namespace will be deleted when the lab pod stops.
        '''
    ).tag(config=True)

    enable_namespace_quotas = Bool(
        True,
        config=True,
        help='''
        If True, will create a ResourceQuota object by calling
        `self.quota_mgr.get_resource_quota_spec()` and create a quota with
        the resulting specification within the namespace.

        A subclass should override the quota manager's
        define_resource_quota_spec() to build a
        situationally-appropriate resource quota spec.
        '''
    ).tag(config=True)

    def __init__(self, *args, **kwargs):
        self.log = make_logger()
        super().__init__(*args, **kwargs)
        self.log.debug("Creating LSSTSpawner.")
        # Our API and our RBAC API are set in the super() __init__()
        # We assume that we're using an LSST Authenticator, which will
        #  therefore have an LSST MiddleManager.
        #
        # This might change with Argo Workflow.
        # self.log.debug("Spawner: {}".format(json.dumps(self.dump())))
        self.log.debug("Setting LSST Manager from authenticated user.")
        user = self.user
        if not user:
            self.log.error("No user found!")
            raise RuntimeError("Could not create spawner!")
        auth = user.authenticator
        if not auth:
            self.log.error("No authenticator found!")
            raise RuntimeError("Could not create spawner!")
        lm = auth.lsst_mgr
        if not lm:
            self.log.error("No LSST Manager found!")
            raise RuntimeError("Could not create spawner!")
        self.lsst_mgr = lm
        lm.spawner = self
        lm.user = user
        self.log.debug("Initialized {}".format(__name__))

    def auth_state_hook(self, spawner, auth_state):
        # Turns out this is in the wrong place.  It should be called
        #  _before_ get_options_form()
        super().auth_state_hook(spawner, auth_state)
        self.log.debug("{} auth_state_hook firing.".format(__name__))

    @gen.coroutine
    def get_options_form(self):
        '''Present an LSST-tailored options form; delegate to options
        form manager.

        This really is stuff that should get set from auth_state_hook...
        But as it happens, that doesn't run until after get_options_form.
        '''
        lm = self.lsst_mgr
        # Take the APIs from the api_manager
        self.api = lm.api_mgr.api
        self.rbac_api = lm.api_mgr.rbac_api
        self.wf_api = lm.api_mgr.wf_api
        om = lm.optionsform_mgr
        auth_state = yield self.user.get_auth_state()
        if not auth_state:
            raise ValueError("Auth state empty for user {}".format(self.user))
        lm.auth_mgr.auth_state = auth_state
        _ = yield self.asynchronize(lm.auth_mgr.parse_auth_state)
        # Will throw error if no UID
        form = yield self.asynchronize(om.get_options_form)
        return form

    def set_user_namespace(self):
        '''Get namespace and store it here (for spawning) and in
        namespace_mgr.'''
        ns = self.get_user_namespace()
        self.namespace = ns
        self.lsst_mgr.namespace_mgr.set_namespace(ns)

    def get_user_namespace(self):
        '''Return namespace for user pods (and ancillary objects).
        '''
        defname = self._namespace_default()
        # We concatenate the default namespace and the name so that we
        #  can continue having multiple Jupyter instances in the same
        #  k8s cluster in different namespaces.  The user namespaces must
        #  themselves be namespaced, as it were.
        if defname == "default":
            raise ValueError("Won't spawn into default namespace!")
        return "{}-{}".format(defname, self.user.escaped_name)

    def start(self):
        # All we need to do is ensure the namespace and K8s ancillary
        #  resources before we run the superclass method to spawn a pod,
        #  so we have the namespace to spawn into, and the service account
        #  with appropriate roles and rolebindings.
        self.log.debug("Starting; creating namespace and ancillary objects.")
        self.set_user_namespace()  # Set namespace here and in namespace_mgr
        self.lsst_mgr.ensure_resources()
        retval = super().start()
        return retval

    @gen.coroutine
    def stop(self, now=False):
        '''After stopping pod, delete the namespace if that option is set.
        '''
        _ = yield super().stop(now)
        if self.delete_namespace_on_stop:
            nsm = self.lsst_mgr.namespace_mgr
            self.log.debug("Attempting to delete namespace.")
            self.asynchronize(nsm.maybe_delete_namespace)
        else:
            self.log.debug("'delete_namespace_on_stop' not set.")

    def options_from_form(self, formdata=None):
        '''Delegate to form manager.
        '''
        # LSST Manager sometimes unset when we make it here.  I don't
        #  understand why that is.  Let's try moving the hooking up
        #  lsst_mgr into __init__ rather than get_form_options.
        return self.lsst_mgr.optionsform_mgr.options_from_form(formdata)

    @gen.coroutine
    def get_pod_manifest(self):
        # Extend pod manifest.  This is a monster method.
        # Run the superclass version, and then extract the fields
        self.log.debug("Creating pod manifest.")
        orig_pod = yield super().get_pod_manifest()
        labels = orig_pod.metadata.labels.copy()
        annotations = orig_pod.metadata.annotations.copy()
        ctrs = orig_pod.spec.containers
        cmd = None
        if ctrs and len(ctrs) > 0:
            cmd = ctrs[0].args or ctrs[0].command
        # That should be it from the standard get_pod_manifest
        # Now we finally need all that data we have been managing.
        # Add label and annotations for ArgoCD management.
        labels['argocd.argoproj.io/instance'] = 'nublado-users'
        annotations['argocd.argoproj.io/compare-options'] = 'IgnoreExtraneous'
        annotations['argocd.argoproj.io/sync-options'] = 'Prune=false'
        cfg = self.lsst_mgr.config
        em = self.lsst_mgr.env_mgr
        am = self.lsst_mgr.auth_mgr
        nm = self.lsst_mgr.namespace_mgr
        vm = self.lsst_mgr.volume_mgr
        om = self.lsst_mgr.optionsform_mgr
        # Get the standard env and then update it with the environment
        # from our environment manager, except that we want the tokens from
        # the standard env
        tokens = {}
        pod_env = self.get_env()
        for fld in ['JUPYTERHUB_API_TOKEN', 'JPY_API_TOKEN']:
            val = pod_env.get(fld)
            if val:
                tokens[fld] = val
        em.create_pod_env()
        pod_env.update(em.get_env())
        pod_env.update(tokens)
        # And now glue in environment information that was stashed at
        #  authentication time (UID/GIDs, authenticator-specific settings)
        #  in the auth_mgr.
        pod_env.update(am.pod_env)
        # Set some constants
        # First pulls can be really slow for the LSST stack containers,
        #  so let's give it a big timeout (this is in seconds)
        self.http_timeout = 60 * 15
        self.start_timeout = 60 * 15
        # We are running the Lab at the far end, not the old Notebook
        self.default_url = '/lab'
        # We always want to check for refreshed images.
        self.image_pull_policy = 'Always'
        # If we can spawn other pods from the Lab, we need a service account.
        self.lab_service_account = nm.service_account
        # Get image name
        pod_name = self.pod_name
        # Get default image name; we will try to replace from options form.
        image = self.image or self.orig_pod.image or cfg.lab_default_image
        # Same with tag.
        tag = "latest"
        # Parse options form result.
        size = None
        image_size = None
        clear_dotlocal = False
        if not self.user_options:
            raise ValueError("No options form data!")
        self.log.debug("User options from form:\n" +
                       json.dumps(self.user_options, sort_keys=True, indent=4))
        size = self.user_options.get('size')
        if not size:
            # Use the deployment defaults
            size_idx = cfg.size_index
            size = cfg.form_sizelist[size_idx]
        image_size = om.sizemap[size]
        self.log.debug("Image size: {}".format(
            json.dumps(image_size, sort_keys=True, indent=4)))
        mem_limit = image_size["mem"]
        cpu_limit = image_size["cpu"]
        mem_guar = image_size["mem_guar"]
        cpu_guar = image_size["cpu_guar"]
        self.mem_guarantee = mem_guar
        self.cpu_guarantee = cpu_guar
        pod_env['MEM_GUARANTEE'] = mem_guar
        pod_env['MEM_LIMIT'] = mem_limit
        pod_env['CPU_GUARANTEE'] = str(cpu_guar)
        pod_env['CPU_LIMIT'] = str(cpu_limit)
        if self.user_options.get('kernel_image'):
            image = self.user_options.get('kernel_image')
            colon = image.find(':')
            if colon > -1:
                imgname = image[:colon]
                tag = image[(colon + 1):]
                if tag == "recommended" or tag.startswith("latest"):
                    # Resolve convenience tags to real build tags.
                    self.log.debug("Resolving tag '{}'".format(tag))
                    qtag = om.resolve_tag(tag)
                    if qtag:
                        tag = qtag
                        image = imgname + ":" + tag
                    else:
                        self.log.warning(
                            "Failed to resolve tag '{}'".format(tag))
                self.log.debug("Image name: %s ; tag: %s" % (imgname, tag))
                if tag == "__custom":
                    self.log.debug("Tag is __custom: retrieving real value " +
                                   "from drop-down list.")
                    cit = self.user_options.get('image_tag')
                    if cit:
                        tag = cit
                        image = imgname + ":" + cit
            self.log.debug("Replacing image from options form: %s" % image)
            self.image = image
        pod_env['JUPYTER_IMAGE_SPEC'] = image
        pod_env['JUPYTER_IMAGE'] = image
        # Set flag to clear .local if indicated
        clear_dotlocal = self.user_options.get('clear_dotlocal')
        if clear_dotlocal:
            pod_env['CLEAR_DOTLOCAL'] = "TRUE"
        # Set up Lab pod resource constraints (not namespace quotas)
        # These are the defaults from the config
        # We don't care about the image name anymore: the user pod will
        #  be named "nb" plus the username and tag, to keep the pod name
        #  short.
        rt_tag = tag.replace('_', '-')
        pn_template = "nb-{username}-" + rt_tag
        pod_name = self._expand_user_properties(pn_template)
        self.pod_name = pod_name
        self.log.debug("Replacing pod name from options form: %s" % pod_name)
        # Get quota definitions from quota manager.
        if self.enable_namespace_quotas:
            qmq = self.lsst_mgr.quota_mgr.quota
            if qmq:
                if "limits.cpu" in qmq:
                    pod_env['NAMESPACE_CPU_LIMIT'] = qmq["limits.cpu"]
                if "limits.memory" in qmq:
                    nmlimit = qmq["limits.memory"]
                    if nmlimit[-2:] == "Mi":
                        nmlimit = nmlimit[:-2] + "M"
                    pod_env['NAMESPACE_MEM_LIMIT'] = nmlimit
        # Get volume definitions from volume manager.
        vm.make_volumes_from_config()
        pod_env['DASK_VOLUME_B64'] = vm.get_dask_volume_b64()
        self.volumes = vm.k8s_volumes
        self.volume_mounts = vm.k8s_vol_mts
        # Add multus annotations if requested
        if cfg.enable_multus:
            annotations.update(
                {'k8s.v1.cni.cncf.io/networks': 'kube-system/macvlan-conf'})
        # Generate the pod definition.
        sanitized_env = sanitize_dict(pod_env, [
            'ACCESS_TOKEN', 'GITHUB_ACCESS_TOKEN',
            'JUPYTERHUB_API_TOKEN', 'JPY_API_TOKEN'])
        self.log.debug("Pod environment: {}".format(json.dumps(sanitized_env,
                                                               sort_keys=True,
                                                               indent=4)))
        self.log.debug("About to run make_pod()")
        pod = make_pod(
            name=self.pod_name,
            cmd=cmd,
            port=self.port,
            image=self.image,
            image_pull_policy=self.image_pull_policy,
            image_pull_secret=self.image_pull_secrets,
            node_selector=self.node_selector,
            run_as_uid=self.uid,
            run_as_gid=self.gid,
            fs_gid=self.fs_gid,
            supplemental_gids=self.supplemental_gids,
            run_privileged=self.privileged,
            env=pod_env,
            volumes=self._expand_all(self.volumes),
            volume_mounts=self._expand_all(self.volume_mounts),
            working_dir=self.working_dir,
            labels=labels,
            annotations=annotations,
            cpu_limit=self.cpu_limit,
            cpu_guarantee=self.cpu_guarantee,
            mem_limit=self.mem_limit,
            mem_guarantee=self.mem_guarantee,
            extra_resource_limits=self.extra_resource_limits,
            extra_resource_guarantees=self.extra_resource_guarantees,
            lifecycle_hooks=self.lifecycle_hooks,
            init_containers=self._expand_all(self.init_containers),
            service_account=self.lab_service_account,
            extra_container_config=self.extra_container_config,
            extra_pod_config=self.extra_pod_config,
            extra_containers=self.extra_containers,
            node_affinity_preferred=self.node_affinity_preferred,
            node_affinity_required=self.node_affinity_required,
            pod_affinity_preferred=self.pod_affinity_preferred,
            pod_affinity_required=self.pod_affinity_required,
            pod_anti_affinity_preferred=self.pod_anti_affinity_preferred,
            pod_anti_affinity_required=self.pod_anti_affinity_required,
            priority_class_name=self.priority_class_name,
            logger=self.log,
        )
        return pod

    def dump(self):
        '''Return dict representation suitable for pretty-printing.'''
        sd = {"namespace": self.namespace,
              "uid": self.uid,
              "gid": self.gid,
              "fs_gid": self.fs_gid,
              "supplemental_gids": self.supplemental_gids,
              "extra_labels": self.extra_labels,
              "extra_annotations": self.extra_annotations,
              "delete_grace_period": self.delete_grace_period,
              "privileged": self.privileged,
              "working_dir": self.working_dir,
              "lifecycle_hooks": self.lifecycle_hooks,
              "init_containers": self.init_containers,
              "lab_service_account": self.lab_service_account,
              "extra_container_config": self.extra_container_config,
              "extra_pod_config": self.extra_pod_config,
              "extra_containers": self.extra_containers,
              "delete_namespace_on_stop": self.delete_namespace_on_stop,
              "enable_namespace_quotas": self.enable_namespace_quotas,
              "lsst_mgr": self.lsst_mgr.dump()}
        return sd



import json
import os
from argo.workflows.sdk import Workflow, template
# This is just a K8s V1Container.
from argo.workflows.sdk.templates import V1Container
from kubernetes.client import V1ResourceRequirements, V1SecurityContext
from kubernetes.client.rest import ApiException
from kubernetes import client
from .. import LoggableChild
from ..utils import list_digest, str_true


class LSSTWorkflowManager(LoggableChild):
    '''This is an LSST Manager Class containing LSST (er, Rubin Observatory)-
    specific logic regarding management of Argo Workflows.
    '''

    cmd_vol = None
    cmd_mt = None
    wf_api = None
    workflow = None
    cfg_map = None
    wf_input = None

    def define_configmap(self, data):
        '''This returns a k8s configmap using the data from the new-workflow
        POST.
        '''
        ni_cmd = data['command']
        idkey = list_digest(ni_cmd)
        cm_name = 'command.{}.json'.format(idkey)
        k8s_vol = client.V1Volume(
            name="noninteractive-command",
            config_map=client.V1ConfigMapVolumeSource(
                name=cm_name
            ),
        )
        k8s_mt = client.V1VolumeMount(
            name="noninteractive-command",
            mount_path="/opt/lsst/software/jupyterlab/noninteractive/command/",
            read_only=True
        )
        self.cmd_vol = k8s_vol
        self.cmd_mt = k8s_mt
        # Now the configmap
        cm_data = {}
        cm_data.update(data)
        del cm_data['image']
        del cm_data['size']
        jd = json.dumps(data, sort_keys=True, indent=4)
        k8s_configmap = client.V1ConfigMap(
            metadata=client.V1ObjectMeta(name=cm_name),
            data={'command.json': json.dumps(data)})
        self.log.debug("Created configmap '{}': {}".format(cm_name, jd))
        self.cfg_map = k8s_configmap

    def define_workflow(self, data):
        '''This is basically our equivalent of get_pod_manifest().
        It creates a dict which we will pass to the workflow template
        engine, which will allow it to create an appropriate workflow.
        '''
        wf_input = {}
        # FIXME Right now we can assume data is of type 'cmd'; we need
        # a little tweaking for 'nb' in that the command will be fixed
        # and the execution parameters will differ.
        cfg = self.parent.config
        em = self.parent.env_mgr
        vm = self.parent.volume_mgr
        am = self.parent.auth_mgr
        om = self.parent.optionsform_mgr
        user = self.parent.parent.authenticator.user
        em_env = em.get_env()
        size_map = om.sizemap.get(data['size'])
        if not size_map:
            # Use default size
            cpu = cfg.tiny_cpu_max * (2 ** cfg.size_index)
            mem = cfg.mb_per_cpu * cpu
            cpu_guar = cpu / (float(cfg.lab_size_range))
            mem_guar = cfg.mb_per_cpu * cpu_guar
            size_map = {
                "cpu": str(cpu),
                "mem": "{}M".format(mem),
                "cpu_guar": str(cpu_guar),
                "mem_guar": "{}M".format(mem_guar)
            }
        self.log.debug(
            "Size '{}' resolves to '{}'".format(data['size'], size_map))
        ml = size_map['mem']
        cl = size_map['cpu']
        mg = size_map['mem_guar']
        cg = size_map['cpu_guar']
        # If we actually have this in the environment, use it (Dask requires
        #  it due to our configuration).  If we do not, then make up something
        #  plausible, but it really doesn't matter if we get it right, because
        #  it's not like we're using the Dask proxy dashboard from inside
        #  a Workflow anyway.
        synth_jsp = '/nb/user/{}'.format(user.escaped_name)
        jsp = os.getenv('JUPYTERHUB_SERVER_PREFIX', synth_jsp)
        wf_input['mem_limit'] = ml
        wf_input['mem_guar'] = mg
        wf_input['cpu_limit'] = str(cl)
        wf_input['cpu_guar'] = str(cg)
        wf_input['image'] = data['image']
        wf_input['enable_multus'] = cfg.enable_multus
        env = {}
        vols = []
        vmts = []
        env.update(em_env)
        vols.extend(vm.k8s_volumes)
        vmts.extend(vm.k8s_vol_mts)
        self.define_configmap(data)
        vols.append(self.cmd_vol)
        vmts.append(self.cmd_mt)
        env['DASK_VOLUME_B64'] = vm.get_dask_volume_b64()
        cname = "wf-{}-{}-{}".format(
            user.escaped_name,
            data['image'].split(':')[-1].replace('_', '-'),
            data['command'][0].split('/')[-1].replace('_', '-'))
        wf_input['name'] = cname
        env['MEM_LIMIT'] = ml
        env['MEM_GUARANTEE'] = mg
        env['CPU_LIMIT'] = str(cl)
        env['CPU_GUARANTEE'] = str(cg)
        env['JUPYTERHUB_USER'] = user.escaped_name
        env['NONINTERACTIVE'] = "TRUE"
        env['EXTERNAL_UID'] = str(user.auth_state['uid'])
        env['EXTERNAL_GROUPS'] = am.get_group_string()
        env['JUPYTERHUB_SERVER_PREFIX'] = jsp
        env['DEBUG'] = str_true(cfg.debug)
        # Get token, if we have one.
        token = self.parent.parent.authenticator.token
        if token:
            env['ACCESS_TOKEN'] = token
        e_l = self._d2l(env)
        wf_input['env'] = e_l
        wf_input['username'] = user.escaped_name
        # Volumes and mounts aren't JSON-serializable...
        wf_input['vols'] = '{}'.format(vols)
        wf_input['vmts'] = '{}'.format(vmts)
        self.wf_input = wf_input
        self.log.debug("Input to Workflow Manipulator: {}".format(
            json.dumps(wf_input, indent=4, sort_keys=True)))
        # ...now put the real values back
        wf_input['vols'] = vols
        wf_input['vmts'] = vmts
        wf_input['command'] = data['command']
        self.wf_input = wf_input
        wf = LSSTWorkflow(parms=wf_input)
        self.log.debug("Workflow: {}".format(wf))
        self.workflow = wf

    def _d2l(self, in_d):
        ll = []
        for k in in_d:
            ll.append({"name": k,
                       "value": in_d[k]})
        return ll

    def list_workflows(self):
        namespace = self.parent.namespace_mgr.namespace
        api = self.parent.wf_api
        self.log.debug(
            "Listing workflows in namespace '{}'".format(namespace))
        nl = self.parent.api.list_namespace(timeout_seconds=1)
        found = False
        for ns in nl.items:
            nsname = ns.metadata.name
            if nsname == namespace:
                self.log.debug("Namespace {} found.".format(namespace))
                found = True
                break
        if not found:
            self.log.debug("No namespace {} found.".format(namespace))
            wfs = None
        else:
            wfs = api.list_namespaced_workflows(namespace=namespace)
        return wfs

    def create_workflow(self):
        workflow = self.workflow
        namespace = self.parent.namespace_mgr.namespace
        api = self.parent.wf_api
        self.create_configmap()
        self.log.debug(
            "Creating workflow in namespace {}: '{}'".format(
                namespace, workflow))
        wf = api.create_namespaced_workflow(namespace, workflow)
        return wf

    def create_configmap(self):
        api = self.parent.api  # Core, not Workflow
        namespace = self.parent.namespace_mgr.namespace
        cfgmap = self.cfg_map
        try:
            self.log.info(
                "Attempting to create configmap in {}".format(namespace))
            api.create_namespaced_config_map(namespace, cfgmap)
        except ApiException as e:
            if e.status != 409:
                estr = "Create configmap failed: {}".format(e)
                self.log.exception(estr)
                raise
            else:
                self.log.info("Configmap already exists.")

    def submit_workflow(self, data):
        self.define_workflow(data)
        nm = self.parent.namespace_mgr
        nm.ensure_namespace()
        nm.ensure_namespaced_service_account()
        wf = self.create_workflow()
        return wf

    def delete_workflow(self, wfid):
        namespace = self.parent.namespace_mgr.namespace
        api = self.parent.wf_api
        wf = api.delete_namespaced_workflow(namespace, wfid)
        return wf

    def get_workflow(self, wfid):
        namespace = self.parent.namespace_mgr.namespace
        api = self.parent.wf_api
        wf = api.get_namespaced_workflow(namespace, wfid)
        return wf

    def dump(self):
        '''Return contents dict for aggregation and pretty-printing.
        '''
        wd = {"parent": str(self.parent),
              "cmd_vol": str(self.cmd_vol),
              "workflow": str(self.workflow),
              "cfg_map": self.cfg_map,
              "wf_input": self.wf_input
              }
        return wd


class LSSTWorkflow(Workflow):
    parms = {}
    entrypoint = "noninteractive"
    run_as_user = 769
    run_as_group = 769

    def __init__(self, *args, **kwargs):
        self.parms = kwargs.pop('parms')
        super().__init__(*args, **kwargs)
        self.spec.volumes = self.parms['vols']
        lbl = {'argocd.argoproj.io/instance': 'nublado-users'}
        self.metadata.labels = lbl
        self.metadata.generate_name = self.parms['name'] + '-'
        self.metadata.name = None
        username = self.parms['username']
        account = "{}-svcacct".format(username)
        self.spec.service_account_name = account
        self.metadata.annotations = self.build_annotations()

    def build_annotations(self):
        anno = {}
        # Add annotations telling Argo CD to not prune these resources or to
        # count them against the sync state.
        anno['argocd.argoproj.io/compare-options'] = 'IgnoreExtraneous'
        anno['argocd.argoproj.io/sync-options'] = 'Prune=false'

        # Enable multus if requested
        if self.parms['enable_multus']:
            anno['k8s.v1.cni.cncf.io/networks'] = 'kube-system/macvlan-conf'

        cmdlist = self.parms['command']
        if not cmdlist:
            return anno
        annobase = "lsst.org/wf_cmd"
        idx = 0
        for cmd in cmdlist:
            akey = annobase + "_{}".format(idx)
            if len(cmd) < 64:
                anno[akey] = cmd
            else:
                jdx = 0
                while cmd:
                    chunk = cmd[:63]
                    cmd = cmd[63:]
                    skey = akey + "_{}".format(jdx)
                    anno[skey] = chunk
                    jdx = jdx+1
            idx = idx + 1

        return anno

    @template
    def noninteractive(self) -> V1Container:
        container = V1Container(
            command=["/opt/lsst/software/jupyterlab/provisionator.bash"],
            args=[],
            image=self.parms["image"],
            name=self.parms["name"],
            env=self.parms["env"],
            image_pull_policy="Always",
            volume_mounts=self.parms["vmts"],
            resources=V1ResourceRequirements(
                limits={"memory": self.parms['mem_limit'],
                        "cpu": self.parms['cpu_limit']},
                requests={"memory": self.parms['mem_guar'],
                          "cpu": self.parms['cpu_guar']}
            ),
            security_context=V1SecurityContext(
                run_as_group=self.run_as_group,
                run_as_user=self.run_as_user,
            )
        )
        self.volumes = self.parms['vols']
        lbl = {'argocd.argoproj.io/instance': 'nublado-users'}
        self.metadata.labels = lbl
        self.metadata.generate_name = self.parms['name'] + '-'
        self.metadata.name = None
        self.service_account_name = self.parms['username'] + '-svcacct'

        return container



'''This is a mixin class for authenticators for common LSST functionality.
'''
import asyncio
from jupyterhub.auth import Authenticator
from .. import LSSTMiddleManager
from ..config import LSSTConfig


class LSSTAuthenticator(Authenticator):
    '''We create an LSST Manager structure on startup; this is an
    LSSTMiddleManager controlling a set of other managers: auth, env,
    namespace, quota, and volume.

    All LSSTAuthenticator subclasses are expected to create two new fields in
     auth_state:
      * auth_state['group_map'], which contains a dict mapping group name
        (the key) to a group ID number (the value).  GIDs may be strings or
        integers.
      * auth_state['uid'], which contains a string or an integer with the
        user's numeric UID.
    '''
    enable_auth_state = True
    delete_invalid_users = True
    token = None  # Only used for JWT, but we want it as a generic field for
    # the workflow manager.

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.lsst_mgr = LSSTMiddleManager(parent=self,
                                          authenticator=self,
                                          config=LSSTConfig())

    def resolve_cilogon(self, membership):
        '''Shared between CILogon and JWT (which uses CILogon as its backing
        store) and thus appearing here.  Returns a uid, groupmap pair
        suitable for insertion into auth_state; both uid and group
        values are strings.
        '''
        am = self.lsst_mgr.auth_mgr
        cfg = self.lsst_mgr.config
        groupmap = {}
        for grp in membership['isMemberOf']:
            name = grp['name']
            gid = grp.get('id')
            if not id and not cfg.strict_ldap_groups:
                gid = am.get_fake_gid()
            if gid:
                groupmap[name] = str(gid)
        uid = str(membership['uidNumber'])
        return uid, groupmap

    def dump(self):
        '''Return dict suitable for pretty-printing.
        '''
        ad = {"enable_auth_state": self.enable_auth_state,
              "delete_invalid_users": self.delete_invalid_users,
              "login_handler": str(self.login_handler),
              "lsst_mgr": self.lsst_mgr.dump()
              }
        return ad

    async def refresh_user(self, user, handler=None):
        '''On each refresh_user, clear the options form cache, thus
        forcing it to be rebuilt on next display.  Otherwise it is built once
        per user session, which is not frequent enough to display new images
        in a timely fashion.
        '''
        self.lsst_mgr.optionsform_mgr.options_form_data = None
        retval = await super().refresh_user(user, handler)
        return retval



# Inspired by, and based on, Adam Tilghman's Multi-Namespace work in
#  https://github.com/jupyterhub/kubespawner/pull/218
import datetime
import pytz
import time
from kubernetes import watch
from kubespawner.reflector import NamespacedResourceReflector
from traitlets import Bool
# This is kubernetes client implementation specific, but we need to know
# whether it was a network or watch timeout.
from urllib3.exceptions import ReadTimeoutError


class MultiNamespaceResourceReflector(NamespacedResourceReflector):
    list_method_omit_namespace = Bool(
        False,
        help="""
        If True, our calls to API `list_method_name` will omit a `namespace`
        argument.  Necessary for non-namespaced methods such as
        `list_pod_for_all_namespaces`
        """
    )

    def _create_resource_key(self, resource):
        """Maps a Kubernetes resource object onto a hashable Dict key;
        subclass may override if `resource.metadata.name` is not
        unique (e.g. pods across multiple namespaces)
        """
        return resource.metadata.name

    def _list_and_update(self):
        """
        Update current list of resources by doing a full fetch.
        Overwrites all current resource info.
        """

        list_args = {
            'label_selector': self.label_selector,
            'field_selector': self.field_selector,
            '_request_timeout': self.request_timeout
        }

        if not self.list_method_omit_namespace:
            list_args['namespace'] = self.namespace

        initial_resources = getattr(
            self.api, self.list_method_name)(**list_args)

        # This is an atomic operation on the dictionary!
        self.resources = {self._create_resource_key(
            p): p for p in initial_resources.items}
        # return the resource version so we can hook up a watch
        return initial_resources.metadata.resource_version

    def _watch_and_update(self):
        """
        Keeps the current list of resources up-to-date

        This method is to be run not on the main thread!

        We first fetch the list of current resources, and store that. Then we
        register to be notified of changes to those resources, and keep our
        local store up-to-date based on these notifications.
        We also perform exponential backoff, giving up after we hit 32s
        wait time. This should protect against network connections dropping
        and intermittent unavailability of the api-server. Every time we
        recover from an exception we also do a full fetch, to pick up
        changes that might've been missed in the time we were not doing
        a watch.

        Note that we're playing a bit with fire here, by updating a dictionary
        in this thread while it is probably being read in another thread
        without using locks! However, dictionary access itself is atomic,
        and as long as we don't try to mutate them (do a 'fetch / modify /
        update' cycle on them), we should be ok!
        """
        selectors = []
        if self.label_selector:
            selectors.append("label selector=%r" % self.label_selector)
        if self.field_selector:
            selectors.append("field selector=%r" % self.field_selector)
        log_selector = ', '.join(selectors)

        cur_delay = 0.1
        ns = self.namespace
        if self.list_method_omit_namespace:
            ns = "[GLOBAL]"

        self.log.info(
            "watching for %s with %s in namespace %s",
            self.kind, log_selector, ns,
        )
        while True:
            start = time.monotonic()
            w = watch.Watch()
            try:
                resource_version = self._list_and_update()
                if not self.first_load_future.done():
                    # signal that we've loaded our initial data
                    self.first_load_future.set_result(None)
                watch_args = {
                    'label_selector': self.label_selector,
                    'field_selector': self.field_selector,
                    'resource_version': resource_version,
                }
                if not self.list_method_omit_namespace:
                    watch_args['namespace'] = self.namespace
                if self.request_timeout:
                    # set network receive timeout
                    watch_args['_request_timeout'] = self.request_timeout
                if self.timeout_seconds:
                    # set watch timeout
                    watch_args['timeout_seconds'] = self.timeout_seconds
                # in case of timeout_seconds, the w.stream just exits
                #  (no exception thrown)
                #     -> we stop the watcher and start a new one
                for ev in w.stream(
                        getattr(self.api, self.list_method_name),
                        **watch_args
                ):
                    cur_delay = 0.1
                    resource = ev['object']
                    if ev['type'] == 'DELETED':
                        # This is an atomic delete operation on the dictionary!
                        self.resources.pop(
                            self._create_resource_key(resource), None)
                    else:
                        # This is an atomic operation on the dictionary!
                        self.resources[self._create_resource_key(
                            resource)] = resource
                    if self._stop_event.is_set():
                        self.log.info("%s watcher stopped", self.kind)
                        break
                        break
                    watch_duration = time.monotonic() - start
                    if watch_duration >= self.restart_seconds:
                        self.log.debug(
                            "Restarting %s watcher after %i seconds",
                            self.kind, watch_duration,
                        )
                        break
            except ReadTimeoutError:
                # network read time out, just continue and restart the watch
                # this could be due to a network problem or just low activity
                self.log.warning(
                    "Read timeout watching %s, reconnecting", self.kind)
                continue
            except Exception:
                cur_delay = cur_delay * 2
                if cur_delay > 30:
                    self.log.exception(
                        "Watching resources never recovered, giving up")
                    if self.on_failure:
                        self.on_failure()
                    return
                self.log.exception(
                    "Error when watching resources, retrying in" +
                    " %ss" % cur_delay)
                time.sleep(cur_delay)
                continue
            else:
                # no events on watch, reconnect
                # This is super-spammy, so don't log it
                # self.log.debug("%s watcher timeout", self.kind)
                pass
            finally:
                w.stop()
                if self._stop_event.is_set():
                    self.log.info("%s watcher stopped", self.kind)
                    break
        self.log.warning("%s watcher finished", self.kind)


class EventReflector(MultiNamespaceResourceReflector):
    kind = 'events'

    list_method_name = 'list_namespaced_event'

    @property
    def events(self):
        return sorted(
            self.resources.values(),
            key=lambda event: event.last_timestamp or event.event_time,
        )


class PodReflector(MultiNamespaceResourceReflector):
    kind = 'pods'
    # FUTURE: These labels are the selection labels for the PodReflector. We
    # might want to support multiple deployments in the same namespace, so we
    # would need to select based on additional labels such as `app` and
    # `release`.
    labels = {
        'component': 'singleuser-server',
    }

    list_method_name = 'list_namespaced_pod'

    @property
    def pods(self):
        return self.resources


class MultiNamespacePodReflector(PodReflector):
    list_method_name = 'list_pod_for_all_namespaces'
    list_method_omit_namespace = True

    def _create_resource_key(self, resource):
        return (resource.metadata.namespace, resource.metadata.name)



#!/usr/bin/env python3
from .scanrepo import ScanRepo
from .parse_args import parse_args


def standalone():
    '''Standalone command for scanning repo.
    '''
    args = parse_args()
    scanner = ScanRepo(host=args.repo,
                       path=args.path,
                       owner=args.owner,
                       name=args.name,
                       dailies=args.dailies,
                       weeklies=args.weeklies,
                       releases=args.releases,
                       experimentals=args.experimentals,
                       recommended=args.recommended,
                       json=args.json,
                       insecure=args.insecure,
                       sort_field=args.sort,
                       cachefile=args.cachefile,
                       debug=args.debug)
    scanner.scan()
    scanner.report()


if __name__ == "__main__":
    standalone()



'''LSST Authentication classes.
'''
from .lsstauth import LSSTAuthenticator
from .lsstcilogonauth import LSSTCILogonOAuthenticator
from .lsstgithubauth import LSSTGitHubOAuthenticator
from .lsstjwtauth import LSSTJWTAuthenticator
from .lsstjwtloginhandler import LSSTJWTLoginHandler

__all__ = [LSSTAuthenticator, LSSTCILogonOAuthenticator,
           LSSTGitHubOAuthenticator, LSSTJWTAuthenticator,
           LSSTJWTLoginHandler]
